#include <GL/glew.h>
#include <stdio.h>
#include <string.h>

#include "state_manager.h"
#include "opengl/shader-gl.h"

ShaderGL::ShaderGL()
{
	sdr = 0;
}

ShaderGL::~ShaderGL()
{
	destroy();
}

bool ShaderGL::create(char *buf, unsigned int bsz, const char *fname)
{
	/* find shader type and create shader */
	unsigned int stype;
	switch(type) {
	case SDR_VERTEX:
		stype = GL_VERTEX_SHADER;
		break;
	case SDR_FRAGMENT:
		stype = GL_FRAGMENT_SHADER;
		break;
	default:
		fprintf(stderr, "Unknown shader type.\n");
		return false;
	}
	name = std::string(fname);
	sdr = glCreateShader(stype);

	/* compile */
	glShaderSource(sdr, 1, (const char **)&buf, 0);
	glCompileShader(sdr);

	delete [] buf;

	/* check compile status */
	int status;
	glGetShaderiv(sdr, GL_COMPILE_STATUS, &status);
	if(status)
		printf("Successfully compiled shader: %s\n", fname);
	else
		fprintf(stderr, "Failed to compile %s shader.\n", fname);

	/* print the info log */
	int loglen;
	glGetShaderiv(sdr, GL_INFO_LOG_LENGTH, &loglen);
	if(loglen > 0 && (buf = new char[loglen + 1])) {
		glGetShaderInfoLog(sdr, loglen, 0, buf);
		buf[loglen] = 0;
		printf("%s\n", buf);

		delete [] buf;
	}

	if(!status) {
		destroy();
		return false;
	}

	return true;
}

void ShaderGL::destroy()
{
	if(sdr) {
		glDeleteShader(sdr);
	}
	sdr = 0;
	type = SDR_UNKNOWN;
}

/* Shader Program */

ShaderProgramGL::ShaderProgramGL()
{
	prog = 0;
	memset(shaders, 0, sizeof shaders / sizeof *shaders);

	is_linked = false;
}

ShaderProgramGL::~ShaderProgramGL()
{
	destroy();
}

bool ShaderProgramGL::create()
{
	prog = glCreateProgram();
	if(!prog) {
		fprintf(stderr, "Failed to create shader program.\n");
		return false;
	}
	return true;
}

bool ShaderProgramGL::link()
{
	if(is_linked)
		return true;

	glLinkProgram(prog);

	int status;
	glGetProgramiv(prog, GL_LINK_STATUS, &status);
	if(status) {
		printf("Successfully linked shader program.\n");
		is_linked = true;
	}
	else
		printf("Failed to link shader program.\n");

	int loglen;
	glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &loglen);

	char *buf;
	if(loglen > 0 && (buf = new char[loglen + 1])) {
		glGetProgramInfoLog(prog, loglen, 0, buf);
		buf[loglen] = 0;
		printf("%s\n", buf);
		delete [] buf;
	}

	return status ? true : false;
}

bool ShaderProgramGL::use() const
{
	if(!is_linked) { //&& !link()) {
		return false;
	}

	if(!prog) {
		return false;
	}

	glUseProgram(prog);
	return true;
}

void ShaderProgramGL::destroy()
{
	glDeleteProgram(prog);
	prog = 0;
	is_linked = false;
}

void ShaderProgramGL::attach_shader(Shader *shader)
{
	glAttachShader(prog, ((ShaderGL *)shader)->sdr);
	is_linked = false;
}

int ShaderProgramGL::get_uniform_location(const char *name) const
{
	if(!use())
		return -1;

	return glGetUniformLocation(prog, name);
}

void ShaderProgramGL::set_uniformi(int location, int value)
{
	if(!use() || location == -1) {
		return;
	}

	glUniform1i(location, value);
}

void ShaderProgramGL::set_uniformi(int location, int x, int y)
{
	if(!use() || location == -1) {
		return;
	}

	glUniform2i(location, x, y);
}

void ShaderProgramGL::set_uniformi(int location, int x, int y, int z)
{
	if(!use() || location == -1) {
		return;
	}

	glUniform3i(location, x, y, z);
}

void ShaderProgramGL::set_uniformi(int location, int x, int y, int z, int w)
{
	if(!use() || location == -1) {
		return;
	}

	glUniform4i(location, x, y, z, w);
}

void ShaderProgramGL::set_uniformf(int location, float value)
{
	if(!use() || location == -1) {
		return;
	}

	glUniform1f(location, value);
}

void ShaderProgramGL::set_uniformf(int location, float x, float y)
{
	if(!use() || location == -1) {
		return;
	}

	glUniform2f(location, x, y);
}

void ShaderProgramGL::set_uniformf(int location, float x, float y, float z)
{
	if(!use() || location == -1) {
		return;
	}

	glUniform3f(location, x, y, z);
}

void ShaderProgramGL::set_uniformf(int location, float x, float y, float z, float w)
{
	if(!use() || location == -1) {
		return;
	}

	glUniform4f(location, x, y, z, w);
}

void ShaderProgramGL::set_uniform_matrix(int location, const Mat4 &mat)
{
	if(!use() || location == -1) {
		fprintf(stderr, "FOO\n");
		return;
	}

	glUniformMatrix4fv(location, 1, GL_FALSE, mat[0]);
}